{  simg -- command line applications for deterministic image processing
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name bgcorrect}
{@description Correct background using a reference image}

{$I directives.pas}
uses simage,getparam,sysutils;

var par                 : tparams;
    cc                  : CSImg_Config;
    ifn,ofn,fs          : TAAStr;
    rfn                 : ansistring;
    frad,ir             : double;
    ff                  : boolean;

    i,j                 : longint;
    ifwhm               : ^double;
    r,rfwhm,dfwhm       : double;
    fi                  : array of TSImg_FPImage;
    fr,fj,fk            : TSImg_FPImage;
    mii                 : array of CSImg_MetaInfo;
    mir                 : CSImg_MetaInfo;
    fa                  : array of double;

begin
    cc:=CSImg_Config.create(SImg_Config_Default);
    with par do
    begin
        init;
        paramusage:='[<infile>]';
        rfn:='';        defparam('<infile>','Reference input (default: stdin)',rfn);
        ifn.size:=0;    defparam('-i','Input file, can be specified multiple times. If more than one input file is given,\\the background of the sum of inputs is corrected. ',ifn);        deffmt('-i','<infile>');
        ofn.size:=0;    defparam('-o','Output file, must be specified as often as -i',ofn);      deffmt('-o','<outfile>');
        frad:=100;      defparam('-r','Background blurring radius. Default: 100',frad);
        ir:=0.25;       defparam('-q','Interpolation ratio. Default and recommended: 25%',ir);
        ff:=false;      defparam('-f','Force to re-calculate FWHMs',ff);
        fs.size:=0;     defparam('-c','Input image factor relative to reference image. Default: auto-detect. If used, must be given for every input image and must not be zero.',fs);
                        defparam('-h','Help',helpf);
        run;
        done;
    end;
    cc.debug(cc.name+': ifn.size='+int2str(ifn.size)+' rfn='+rfn+' ofn.size='+int2str(ifn.size)+' frad='+real2str(frad)+' ir='+real2str(ir)+' ff='+boolean2str(ff)+' fs.size='+int2str(fs.size));
    if ofn.size=0 then
    begin
        cc.warning(cc.name+': No output defined -- nothing to do');
        SImg_Halt;
    end;
    if ifn.size=0 then
    begin
        cc.warning(cc.name+': No input defined -- nothing to do');
        SImg_Halt;
    end;
    if (ifn.size<>ofn.size) then cc.error(cc.name+': Number of input images must be equal to number of output images');

    { input image factors }
    setlength(fa,ifn.size);
    if fs.size<>0 then
    begin
        if fs.size<>ifn.size then cc.error(cc.name+': Number of input images factors must be equal to number of input image');
        for i:=0 to ifn.size-1 do
        begin
            if not val2(fs.aa[i],fa[i]) then cc.error(cc.name+': Number expected after -c, got "'+fs.aa[i]+'"');
            if abs(fa[i])<simg_mina then cc.error(cc.name+': Input image factor (-c) must not be zero');
        end
    end
    else
        for i:=0 to ifn.size-1 do
            fa[i]:=1;

    { inputs }
    getmem(ifwhm,ifn.size*sizeof(double));
    setlength(fi,ifn.size);
    setlength(mii,ifn.size);
    for i:=0 to ifn.size-1 do
    begin
        fi[i].pt:=nil;
        mii[i]:=CSImg_MetaInfo.create();
        cc.Info(cc.name+': Reading input image '+ifn.aa[i]);
        SImg_ReadImageFile(cc, ifn.aa[i], fi[i], mii[i]);
        if (fi[i].nx<>fi[0].nx) or (fi[i].ny<>fi[0].ny) or (fi[i].nc<>1)  then cc.error(cc.name+': Invalid dimensions of input image'+ int2str(i+1) +' (' + ifn.aa[i] + ')');
        if ff or (not mii[i].get('FWHMPIX',ifwhm[i])) then ifwhm[i]:=SImg_GetSharpnessFWHM(cc,fi[i]);
    end;

    { reference image }
    fr.pt:=nil;
    rfwhm:=ifwhm[0];
    if rfn<>'' then
    begin
        cc.Info(cc.name+': Reading reference image '+rfn);
        mir:=CSImg_MetaInfo.create();
        SImg_ReadImageFile(cc, rfn, fr, mir);
        if (fr.nc<>fi[0].nc) or (abs(fr.nx*fi[0].ny/(fi[0].nx*fr.ny)-1)>0.1) then cc.error(cc.name+': Invalid dimensions of reference image');
        if ff or (not mir.get('FWHMPIX',rfwhm)) then rfwhm:=-1;
        r:=1.25;
        if (fr.nx<>fi[0].nx) or (fr.ny<>fi[0].ny) then
        begin
            r:=sqrt(fi[0].nx*fi[0].ny/(fr.nx*fr.ny));
            cc.Info(cc.name+': Re-Scaling reference image by factor ' +real2str(r,-5));
            SImg_Resize(cc, fr,fr, fr.nx/fi[0].nx, fr.ny/fi[0].ny, -1);
            rfwhm:=-1;
        end;
        if rfwhm<simg_mina then
        begin
            if r<1 then r:=1;
            rfwhm:=SImg_GetSharpnessFWHM(cc,fr, r,round(r*20));
        end;
    end;

    { destination fwhm }
    dfwhm:=rfwhm;
    for i:=0 to ifn.size-1 do
        if ifwhm[i]>dfwhm then dfwhm:=ifwhm[i];

    { filtering }
    if fr.pt<>nil then
    begin
        r:=sqr(dfwhm)-sqr(rfwhm);
        if r>0.1 then
        begin
            r:=sqrt(r);
            cc.Info(cc.name+': Blurring reference image from FWHM='+real2str(rfwhm,-1)+' to FWHM='+real2str(dfwhm,-1));
            SImg_Filter(cc, fr,fr, r*FWHM_TO_RAD);
        end;
    end;
    fj.pt:=nil; fk.pt:=nil;
    for i:=0 to ifn.size-1 do
    begin
        r:=sqr(dfwhm)-sqr(ifwhm[i]);
        if r>0.1 then
        begin
            r:=sqrt(r);
            cc.Info(cc.name+': Blurring input image '+ifn.aa[i]+' from FWHM='+real2str(ifwhm[i],-1)+' to FWHM='+real2str(dfwhm,-1));
            SImg_Filter(cc, fi[i],fk, r*FWHM_TO_RAD);
            if ifn.size=1 then SImg_Move(cc, fk,fj);
        end
        else if ifn.size=1 then SImg_Duplicate(cc, fi[i],fj, true)
        else fk:=fi[i];
        if ifn.size>1 then
        begin
            if fj.pt=nil then SImg_New(cc, fj, fk.nx,fk.ny,fk.nc, false);
            r:=fa[i]/ifn.size;
            for j:=0 to fj.size-1 do
                fj.pt[j]+=r*fk.pt[j];
        end;
        if fk.pt<>fi[i].pt then SImg_free(cc,fk);
        fk.pt:=nil;
    end;

    { background estimation }
    if fs.size=0 then r:=0
    else r:=fa[0];
    cc.Info(cc.name+': Background correction using '+int2str(cc.threads)+' threads');
    SImg_BGCorrect(cc, fj,fr, fj, dfwhm,ir,frad,r);
    SImg_Free(cc, fr);

    { output }
    for i:=0 to ofn.size-1 do
    begin
        fk:=fi[i];
        if ifn.size=1 then r:=1
        else r:=1/fa[i];
        for j:=0 to fj.size-1 do
            fk.pt[j]-=fj.pt[j];
        mii[i].add('FWHMPIX',real2str(ifwhm[i],-4), 'Star FWHM in pixels');
        cc.Info(cc.name+': Writing output image '+ofn.aa[i]);
        SImg_WriteImageFile(cc, ofn.aa[i], fk, mii[i]);
        SImg_Free(cc, fk);
        mii[i].destroy;
    end;
    SImg_Free(cc, fj);
end.
