{  simg -- command line applications for deterministic image processing
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name sdbadd}
{@description Adds one or multiple images to stack database}

{$I directives.pas}
uses simage,getparam,sysutils;

const unmask_bncorr = 1;        { correction of minimum square background noise }
      mosaic_approx = 2;        { boundary overweight in mosaic mode }
//      mosaic_approx = 1;      { boundary overweight in mosaic mode }

var par                         : tparams;
    cc                          : CSImg_Config;
    sfn,ofn,flfn,rfn,cs,cso     : ansistring;
    wb,dt,dr,we,bw              : double;
    fmfn,afn,wfn,ifn,bfn,pfn    : TAAStr;
    bg,sa,c                     : longint;
    r1,tm                       : boolean;

    fs,ffl,fr,fi,fw,ffm,fj,fk   : TSImg_FPImage;
    fa,fb,fp                    : TSImg_FPImage;
    sl                          : TSImg_RDImage;
    res                         : TSImg_AlignmentParams;
    ms,ma,mi,mw,mf,msl          : CSImg_MetaInfo;
    si                          : TSImg_SDBInfo;
    i,j,x,y                     : longint;
    s,t,u,adr,sbn               : double;
    dtd                         : boolean;
    bi                          : TSImg_BayerInfo;
    ps,ca,pbfn,ppfn             : ansistring;
    wcs                         : TSImg_WCS;
    bbw,bbw2                    : ^double;      { buffer for reference source list output }

begin
{ define sdbadd_debug}
{ define sdbadd_noweight}
    cc:=CSImg_Config.create(SImg_Config_Default);
    with par do
    begin
        init;
        paramusage:='<stack database> [<FWHM limits>]';
        sfn:='';        defparam('<stack database>','Stack database',@sfn,tpastr,rapdreq);
        flfn:='';       defparam('<FWHM limits>','FWHM limit image',flfn);
        ifn.size:=0;    defparam('-i','Input image, can be specified multiple times',ifn);      deffmt('-i','<image>');
        afn.size:=0;    defparam('-a','Alignment data. must be defined as often as -i',afn);                                                            deffmt('-a','<image>');
        wfn.size:=0;    defparam('-w','Pixel weight image.  Optional. If defined, as oftan as -i.',wfn);                                                        deffmt('-w','<image>');
        bg:=1;          defparam('-b','Background subtraction (0:none, 1:constant, 2:gradient, 3:curved gradient, 4+:1+ radial polynomials; default:1)',bg);
        bfn.size:=0;    defparam('-bi','Background approximation image. Optional. If defined, as oftan as -i.',bfn);                                            deffmt('-bi','<image>');
        pfn.size:=0;    defparam('-bp','Polynomial approximation image. Optional. If defined, as oftan as -i. Only recommended with reference image.\\'+
                                       'If defined, multiplied with polynomials defined by -b such that DOF are doubled.',pfn);                                 deffmt('-bp','<image>');
        fmfn.size:=0;   defparam('-f','FWHM map, can be specified multiple times',fmfn);        deffmt('-f','<image>');
        rfn:='';        defparam('-r','Reference image for 2nd and later pass or inaccurate reference image for 1st pass (in combination with -r1).\\'+
                                       'Reference image should be background corrected because it is used for estimation of photon noise.',rfn);                deffmt('-r','<image>');
        r1:=false;      defparam('-r1','Enforce pass 1 iteration',r1);
        wb:=-1;         defparam('-wb','Underweight boundaries. Parameter is relative to half of the smaller image image dimensions, i.e. range is 0..1.',wb);
        cs:='';         defparam('-cs','Factor calibration using reference source list which is only recommended if pixel weight is used',cs);                  deffmt('-cs','<source list>');
        cso:='';        defparam('-cso','Output corrected reference source list for next pass.',cso);                                                           deffmt('-cso','<source list>');
        dt:=-1;         defparam('-u','Mask out unlikely pixels in 2nd and later passes. Parameter is the threshold as noise factor. Feasible > 1',dt);
        dr:=1.5;        defparam('-ur','Absolute value is the unmask threshold relative to reference. If <0 only brighter pixel are unmasked.\\'+
                                       'This is not recommended for the last pass. Default: 1.5',dr);
        sa:=2;          defparam('-s','Area brightness compensation. 0:none, 1:WCS, 2:determinant of alignment matrix. Default 2.\\' +
                                      'If brightness is calibrated using stars, the scaling factor includes everything but\\' +
                                      '(angular) pixel area. Due to flatfield corrections the correct correction factor is\\' +
                                      '<pixel area of reference image in center> / <pixel area of destination image in center>\\' +
                                      'Numerator is normalized to 1"^2 for method 1. Method 2 is more or less inaccurate due to\\'+
                                      'distortions. If WCS data is available, method 1 should be used.',sa);
        we:=0;          defparam('-we','Brightness weight exponent, feasible: 0..1. Should be 0 for final pass at max. SNR. Default: 0',we);
        c:=1;           defparam('-c','Use this color if Bayer patter image is recognized. Default: 1 which is the 2nd color',c);
        bw:=0.1;        defparam('-bw','Weight factor for undefined pixels of Bayer pattern images. Default: 0.1', bw);
        tm:=false;      defparam('-m','True mosaic mode: Suppress underweighting boundaries due to flatfield correction',tm);
        ofn:='';        defparam('-o','Output stack database (default: same as input)',ofn);                                                                    deffmt('-o','<outfile>');
                        defparam('-h','Help',helpf);
        run;
        done;
    end;
    cc.debug(cc.name+': sfn='+sfn+' flfn='+flfn+' rfn='+rfn+' ofn='+ofn+' afn.size='+int2str(afn.size)+' ifn.size='+int2str(ifn.size)+' wfn.size='+int2str(wfn.size)+' bfn.size='+int2str(bfn.size)+' pfn.size='+int2str(pfn.size)+' fmfn.size='+int2str(fmfn.size)+
        ' bg='+int2str(bg)+' wbr='+real2str(wb)+' dt='+real2str(dt)+' dr='+real2str(dr)+' cs='+cs+' cso='+cso+' sa='+int2str(sa)+' we='+real2str(we)+' c='+int2str(c)+' bw='+real2str(bw)+' r1='+boolean2str(r1)+' tm='+boolean2str(tm));
    if wb>1 then wb:=1;

    if afn.size<>ifn.size then cc.error(cc.name+': Number of alignment data sets must be equal to number if input images');
    if ifn.size<1 then
    begin
        cc.warning('No input defined -- nothing to do');
        SImg_Halt;
    end;
    if (wfn.size<>0) and (wfn.size<>ifn.size) then cc.error(cc.name+': Number of weight images must be equal to number if input images');
    if (bfn.size<>0) and (bfn.size<>ifn.size) then cc.error(cc.name+': Number of background approximation images must be equal to number if input images');
    if (pfn.size<>0) and (bg<1) then
    begin
        cc.warning(cc.name+': No polynomials defined (parameter -b). Polynomial approximation images are ignored');
        pfn.size:=0;
    end;
    if (pfn.size<>0) and (pfn.size<>ifn.size) then cc.error(cc.name+': Number of flat fields must be equal to number if input images');
    if (flfn='') and (fmfn.size<>0) then
    begin
        fmfn.size:=0;
    end;
    if (flfn<>'') and (fmfn.size=0) then
    begin
        cc.error(cc.name+': No FWHM maps given, ignoring FWHM limit');
        flfn:='';
    end;
    if (fmfn.size<>0) and (fmfn.size<>ifn.size) then cc.error(cc.name+': Number FWHM maps must be equal to number if input images');
    dtd:=dt>0.1;

    { factor calibration initialization }
    sl.pt:=nil;  bbw:=nil;  bbw2:=nil;
    msl:=CSImg_MetaInfo.create();
    if (cso<>'') and (cs='') then cc.error(cc.name+': Output reference source list (parameter -cso) requires input reference source list (parameter -cs)');
    if cs<>'' then
    begin
        cc.Info(cc.name+': Reading reference source list '+cs);
        SImg_ReadImageFile(cc,cs,sl,msl);
        if not (SImg_Recover(cc,sl) and (sl.fmt=SImg_RawDouble) and (sl.nx>=5) and (sl.nc=1)) then cc.error(cc.name+': Invalid reference source list');
    end;
    if cso<>'' then
    begin
        getmem2(bbw,sl.ny*3*sizeof(double));
        getmem2(bbw2,sl.ny*3*sizeof(double));
        for i:=0 to sl.ny*3-1 do
            bbw[i]:=0;
    end;

    { stack database }
    ms:=CSImg_MetaInfo.create();
    fs.pt:=nil;
    cc.Info(cc.name+': Reading stack database '+sfn);
    SImg_ReadImageFile(cc,sfn,fs,ms);
    if (fs.nc and 1)<>0 then cc.error(cc.name+': Invalid stack database: '+sfn);
    SImg_StackDB_GetInfo(cc, si, ms, fs.nx,fs.ny);
    if fs.nc<4 then
    begin
        if (fmfn.size>0) or (flfn<>'') then cc.warning(cc.name+': Stack database contains less than 2 FWHM ranges, ignoring FWHM parameters');
        fmfn.size:=0;
        flfn:='';
    end;

    { reference image }
    fr.pt:=nil;
    if rfn<>'' then
    begin
        cc.Info(cc.name+': Reading reference image '+rfn);
        SImg_ReadImageFile(cc,rfn,fr);
        if (fr.nc<>1) or (abs(fr.nx*fs.ny/(fs.nx*fr.ny)-1)>0.1) then cc.error(cc.name+': Invalid dimensions of reference image');
        if (fr.nx<>fs.nx) or (fr.ny<>fs.ny) then
        begin
            cc.Info(cc.name+': Scaling reference image by factor ' +real2str(sqrt(fs.nx*fs.ny/(fr.nx*fr.ny)),-5));
            SImg_Resize(cc, fr,fr, fr.nx/fs.nx, fr.ny/fs.ny, -1);
        end;
    end;

    { fwhm limits }
    ffl.pt:=nil;
    if flfn<>'' then
    begin
        cc.Info(cc.name+': Reading FWHM limits '+flfn);
        SImg_ReadImageFile(cc,flfn,ffl);
    end;

    { input images }
    ma:=CSImg_MetaInfo.create();
    mi:=CSImg_MetaInfo.create();
    mw:=CSImg_MetaInfo.create();
    mf:=CSImg_MetaInfo.create();
    fi.pt:=nil;  ffm.pt:=nil;
    fw.pt:=nil;  fw.nx:=0;
    fb.pt:=nil;  pbfn:='';
    fp.pt:=nil;  ppfn:='';
    for i:=0 to 3 do
        bi.cf[i]:=1;
//    for i:=0 to 10 do
    for i:=0 to ifn.size-1 do
    begin
        { alignment data }
        cc.Info(cc.name+': Reading alignment data '+afn.aa[i]);
        ma.reset;
        SImg_RHeader(cc,afn.aa[i],ma);
        if not SImg_ReadAlignmentParams(cc,res,ma) then cc.error(cc.name+': Align parameters not set');

        { input image }
        cc.Info(cc.name+': Reading input image '+ifn.aa[i]);
        mi.reset;
        bi.pattern[0,0]:=-1;    // means no Bayer pattern
        SImg_ReadImageFile(cc,ifn.aa[i],fi,mi);
        if fi.nc<>1 then cc.Info(cc.name+': input image contains more than one color');
        if mi.get('BAYERPAT',ps) then
        begin
            cc.Info(cc.name+': Interpolate color ' + int2str(c));
            bi.pattern:=SImg_BayerParsePattern(cc,trimquotes(ps));
            if (sl.pt<>nil) and (bw<0.699) then j:=0                            { in calibration mode non-linear interpolation is only exact if bw=1 }
            else if (fr.pt<>nil) and (dr<=-simg_mina) and (bw<0.101) then j:=-1 { q25 interpolation in unsymmetrical mode }
            else j:=1;                                                          { otherwise median }
            SImg_BayerInterpolation(cc,bi, fi,fi, j, c);
{$ifdef sdbadd_debug}
            if i=0 then SImg_WriteImageFile(cc,'sdbadd-bayer-i.fits',fi);
{$endif}
        end;
        adr:=0;
        if mi.get('EGAIN',adr) and (adr>simg_mina) then adr:=1/adr
        else adr:=0;

        { fwhm map }
        mf.reset;
        if fmfn.size>0 then
        begin
            cc.Info(cc.name+': Reading FWHM map '+fmfn.aa[i]);
            SImg_ReadImageFile(cc,fmfn.aa[i],ffm,mf);
        end;

        { pixel weight }
        if wfn.size>0 then
        begin
            cc.Info(cc.name+': Reading pixel weight image '+wfn.aa[i]);
            mw.reset;
            SImg_ReadImageFile(cc,wfn.aa[i],fw,mw);

            if (fw.nx<>fi.nx) or (fw.ny<>fi.ny) or (fw.nc<>1) then cc.error(cc.name+': Invalid dimensions of weight image '+wfn.aa[i]);
            if bi.pattern[0,0]>=0 then
            begin
                SImg_BayerInterpolation(cc, bi, fw,fw, -2, c);
{$ifdef sdbadd_debug}
                if i=0 then SImg_WriteImageFile(cc,'sdbadd-bayer-w.fits',fw);
{$endif}
            end;
        end
        else begin
            if (fw.nx<>fi.nx) or (fw.ny<>fi.ny) then SImg_New(cc, fw, fi.nx,fi.ny,1, false);
            { try to find saturated pixels }
            s:=fi.pt[0];
            for j:=0 to fi.size-1 do
                if fi.pt[j]>s then s:=fi.pt[j];
            s-=abs(s)*0.25;
            for j:=0 to fw.size-1 do
                if fi.pt[j]<s then fw.pt[j]:=1
                else fw.pt[j]:=0;
        end;

        { polynomial image }
        if pfn.size>0 then
        begin
            if (fp.pt=nil) or (pfn.aa[i]<>ppfn) then
            begin
                cc.Info(cc.name+': Reading polynomial approximation image '+pfn.aa[i]);
                ppfn:=pfn.aa[i];
                SImg_ReadImageFile(cc, ppfn, fp);
                if bi.pattern[0,0]>=0 then SImg_BayerInterpolation(cc, bi, fp,fp, 0, c);
            end;
        end;

        { background approxiamtion image }
        if bfn.size>0 then
        begin
            if (bfn.aa[i]=ppfn) then
            begin
                cc.Info(cc.name+': Warning: Background approximation image equal to polynomial approximation image. Ignoring the background approximation image');
                SImg_Free(cc,fb);
                pbfn:='';
            end
            else if (fb.pt=nil) or (bfn.aa[i]<>pbfn) then
            begin
                cc.Info(cc.name+': Reading background approximation image '+bfn.aa[i]);
                pbfn:=bfn.aa[i];
                SImg_ReadImageFile(cc, pbfn, fb);
                if bi.pattern[0,0]>=0 then SImg_BayerInterpolation(cc, bi, fb,fb, 0, c);
            end;
        end;

        { calibration using reference source list }
        if sl.pt<>nil then
        begin
            if mi.get('FWHMPIX',s) or mf.get('FWHMPIX',s) then cc.Info(cc.name+': Read PSF FWHM width from meta data: ' +real2str(s,-1))
            else begin
                s:=SImg_GetSharpnessFWHM(cc,fi);
                cc.Info(cc.name+': Auto-detected FWHM: ' +real2str(s,-1));
            end;
            s:=SImg_StackDB_calib(cc, fi,fw, sl, res, s, bbw2);
            if s<simg_mina then cc.warning(cc.name+': Could not determine image factor')
            else begin
                cc.Info(cc.name+': Image factor adjusted to '+real2str(s,5)+' (was '+real2str(res.fact,5)+select(res.fact>simg_mina,'  quotient: '+real2str(s/res.fact,5),'')+')');
                res.fact:=s;
            end;
            { update data for output reference list }
            if (bbw<>nil) and (res.fact>simg_mina) then
            begin
                s:=1/res.fact;
                for j:=0 to sl.ny-1 do
                begin
                    x:=j*3;
                    t:=bbw2[x+2]*s;         { weight }
                    bbw[x  ]+=bbw2[x  ]*t;  { brightness }
                    bbw[x+1]+=bbw2[x+1]*t;  { background }
                    bbw[x+2]+=s*t;
                end;
            end;
        end;

        { weight exponent }
        if res.fact>=simg_mina then res.weight*=exp(ln(res.fact)*we);

        { area correction }
        if sa=1 then
        begin
            if not SImg_GetWCSData(cc, ma, wcs) then cc.error(cc.name+': WCS data required for area brightness compensation method 1');
            SImg_GetPixelScale(wcs, s,t);
            res.fact/=abs(s*t);
        end
        else if sa=2 then
        begin
            s:=abs(res.b11*res.b22-res.b12*res.b21);
            res.fact*=s;
            if abs(s-1)>1e-3 then cc.info(cc.name+': Area correction factor: '+real2str(s));
        end;

        { shift reference image and add photon noise to 1/fw }
        fa.pt:=nil;
        if (fr.pt<>nil) and ((bg>0) or dtd) then
        begin
            if (wfn.size>0) and (adr>simg_mina) then
            begin
                s:=adr;
                cc.Info(cc.name+': Shift reference image and estimate photon noise using '+int2str(cc.threads) + ' threads');
            end
            else begin
                s:=0;
                cc.Info(cc.name+': Shift reference image using '+int2str(cc.threads) + ' threads');
            end;
            SImg_New(cc, fa, fi.nx,fi.ny,1);    // undefined pixels will be 0
            if not SImg_StackDB_shift(cc, si, fr,fa,fw, res, s) then
            begin
                SImg_Free(cc, fa);
                continue;
            end;
{$ifdef sdbadd_debug}
            if i=0 then SImg_WriteImageFile(cc,'sdbadd-fa.fits',fa);
{$endif}
        end;

{$ifdef sdbadd_noweight}
        s:=0; t:=0;
        for j:=0 to fw.size-1 do
        begin
            u:=fw.pt[j];
            s+=sqr(u);
            t+=u;
        end;
        s:=0.1*s/t;
        for j:=0 to fw.size-1 do
            if fw.pt[j]>s then fw.pt[j]:=s;
        tm:=false;
{$endif}

        { background removal }
        if (bg>0) or dtd then
        begin
            if (fa.pt<>nil) and r1 then         { pass 1 with reference image }
            begin
                cc.Info(cc.name+': Gradient and/or background removal with reference image using '+int2str(cc.threads) + ' threads');
                SImg_ImgBGApprox(cc, fi,fa,fb,fp,fw, bg, tm,true);
                SImg_Free(cc,fa);
            end
            else if fa.pt<>nil then             { pass >1 with reference image }
            begin
                { estimate square background noise sbn }
                sbn:=0;
                if ((bi.pattern[0,0]<0) or (adr<simg_mina)) and mi.get('BGNOISE',sbn) then sbn:=sqr(sbn)        { enforce re-calculation of background noise for bayer pattern images }
                else if adr>=simg_mina then
                begin
{                   if mi.get('BGNOISE',s) then cc.info('Read background noise: '+real2str(s)+' image units = '+real2str(s/adr)+' e-');
                    if mi.get('RDNOISE',s) then cc.info('Read readout noise: '+real2str(s)+' image units = '+real2str(s/adr)+' e-');
                    if mi.get('DCNOISE',s) then cc.info('Dark current noise: '+real2str(s)+' image units = '+real2str(s/adr)+' e-'); }
                    { photon noise }
                    s:=SImg_GetQuantile(cc, fi, 0.45)*adr;      { background photon noise in image units }
{                   cc.info('Estimated backgrond photon noise: '+real2str(sqrt(s))+' image units = '+real2str(sqrt(s)/adr)+' e-'); }
                    if s<0 then s:=0;
                    sbn+=s;
                    { read out noise }
                    if not mi.get('RDNOISE',t) then
                    begin
                        t:=sqrt(0.5*s);
                        cc.warning(cc.name+': Read out noise not defined, assuming '+real2str(t)+' image units = '+real2str(t/adr)+' e-');
                    end;
                    sbn+=sqr(t);
                    { dark current noise }
                    if not mi.get('DCNOISE',t) then
                    begin
                        t:=sqrt(0.5*s);
                        cc.warning(cc.name+': Dark current noise not defined, assuming '+real2str(t)+' image units = '+real2str(t/adr)+' e-');
                    end;
                    sbn+=sqr(t);
                    t:=sqrt(sbn);
                    cc.info(cc.name+': Estimated background noise: '+real2str(t)+' image units = '+real2str(t/adr)+' e-');
                end
                else begin
                    if mi.get('RDNOISE',t) then sbn+=sqr(t)
                    else cc.warning('Read out noise not defined.');
                    if mi.get('DCNOISE',t) then sbn+=sqr(t)
                    else cc.warning('Dark current noise not defined.');
                    sbn*=2;
                    cc.warning(cc.name+': Gain (ADC resolution) not defined. Unable to correctly estimate background noise. Assuming: '+real2str(sqrt(t))+' image units');
                end;

                { background removal }
                fk.pt:=nil;
                SImg_ImgApprox(cc, fi,fa,fb,fp,fw, fk,fj, bg, tm);  { fk: residual, fj: corrected image }
{$ifdef sdbadd_debug}
                if i=0 then SImg_WriteImageFile(cc,'sdbadd-fres.fits',fk);
{$endif}

                { mask out unlikely values. fi: estimated square noise}
                ca:='';  { makes the compiler happy }
                if dtd and (wfn.size>0) then
                begin
                    { Estimates variance of pixel using weight image and photon noise estimated by SImg_StackDB_shift. If photon noise cannot be estimated it should not be a problem if -ur is large (>1). }
                    ca:='Mask out unlikely values. Noise noise estimation using weight and reference image';
                    s:=sbn*unmask_bncorr;
                    for j:=0 to fi.size-1 do
                    begin
                        t:=fw.pt[j];
                        if t>simg_mina then
                        begin
                            t:=1/t;
                            u:=fi.pt[j]-fk.pt[j]-fa.pt[j];
                            if u>0 then t+=u*adr;                       { use r*fa instead of fa }
                            fi.pt[j]:=t+s;
                        end
                        else fi.pt[j]:=simg_large;
                    end;
                end
                else if dtd and (adr>simg_mina) then
                begin
                    { It is assumed that the background noise contains photon noise of background, dark current noise and read out noise, i.e. variance can be estimated from background noise and fr2.
                      This algorithm does not take into account noise caused by pixel defects, amplifier glow, scaling (vignetting), ... and is therefore not optimal. }
                    ca:='Mask out unlikely values. Noise estimation using reference image, background, gain and readout noise';
                    s:=sbn*unmask_bncorr;
                    for j:=0 to fi.size-1 do
                    begin
                        t:=sbn + (fi.pt[j]-fk.pt[j])*adr;
                        if t<0 then t:=0;
                        fi.pt[j]:=t + s;
                    end;
                end
                else begin
                    if dtd then cc.warning('Unable to estimate noise, usable unmasking');
                    SImg_Free(cc,fi);
                    fi.pt:=nil;
                end;

                if fi.pt=nil then SImg_Move(cc,fj,fi)
                else begin
                    s:=SImg_StackDB_unmask(cc, fj,fk,fi,fw, dt,dr,adr);
                    cc.info(cc.name+': '+ca+': '+real2str(s/fw.size*100,-2)+'%');
                    SImg_Move(cc,fj,fi);
                end;
                SImg_Free(cc,fk);

                { final background approximation with boundary overweight }
                for j:=0 to fi.size-1 do
                    fi.pt[j]-=fa.pt[j];
                SImg_ApproxMosaic(cc, fi,fb,fp,fw, bg, mosaic_approx, tm);
                for j:=0 to fi.size-1 do
                    fi.pt[j]+=fa.pt[j];
                SImg_Free(cc,fa);
                if fj.pt<>nil then SImg_free(cc,fj);
            end
            else begin          { pass 1 / no reference image }
                cc.Info(cc.name+': Gradient and/or background removal '+int2str(cc.threads) + ' threads');
                SImg_BGApprox3(cc, fi,fb,fp,fw, bg, tm);
            end;
        end;

        { underweight boundaries }
{$ifndef sdbadd_noweight}
        if wb>1e-3 then SImg_UnderweightBoundaries(cc,fw,wb);
{$endif}

        { underweight undefined pixels in Bayer images }
        if (bi.pattern[0,0]>=0) and (bw<=0.999) then
        begin
            j:=0;
            for y:=0 to fi.ny-1 do
                for x:=0 to fi.nx-1 do
                begin
                    if bi.pattern[y and 1, x and 1]<>c then fw.pt[j]*=bw;
                    j+=1;
                end;
        end;

{$ifdef sdbadd_debug}
        if i=0 then SImg_WriteImageFile(cc,'sdbadd-fw.fits',fw);
{$endif}

        { add }
        cc.Info(cc.name+': Add image to stack database using '+int2str(cc.threads) + ' threads');
        SImg_StackDB_add(cc, si, fs,fi,fw, ffl,ffm, res);

//    if i=0 then cc.error(cc.name+': No error');
    end;
    SImg_Free(cc,fi);
    SImg_Free(cc,fw);
    SImg_Free(cc,ffl);
    SImg_Free(cc,ffm);
    SImg_Free(cc,fb);
    SImg_Free(cc,fp);

    { adjust factor }
    if fr.pt<>nil then
    begin
        SImg_StackDB_ConvertValWeight(cc, fs, fi,fw);
        s:=SImg_ImgApprox2(cc, fr, fi,fw);
        if (s<0.95) or (s>2) then cc.warning('Invalid correction factor: '+real2str(s,-3)+': Omit correction')
        else begin
            cc.info(cc.name+': Correction factor: '+real2str(s,-3));
            for i:=0 to fs.size div 2-1 do
                fs.pt[i*2]*=s;
        end;
    end;
    SImg_Free(cc,fr);
    SImg_Free(cc,fi);
    SImg_Free(cc,fw);

    { output }
    if ofn='' then ofn:=sfn;
    cc.Info(cc.name+': Write stack database '+ofn);
    SImg_WriteImageFile(cc,ofn,fs,ms);
    SImg_Free(cc,fs);

    { corrected reference source list }
    if cso<>'' then
    begin
        cc.Info(cc.name+': Write corrected reference source list '+cso);
        for i:=0 to sl.ny-1 do
        begin
            x:=i*3;
            s:=bbw[x+2];
            if s>simg_mina then
            begin
                s:=1/s;
                y:=i*sl.sl;
                sl.ptd[y+2]:=bbw[x  ]*s;        { brightness }
                sl.ptd[y+3]:=bbw[x+1]*s;        { background }
                sl.ptd[y+4]:=1;
            end;
        end;
        freemem2(bbw,sl.ny*3*sizeof(double));
        freemem2(bbw2,sl.ny*3*sizeof(double));
        SImg_WriteImageFile(cc,cso,sl,msl);
    end;
    SImg_Free(cc, sl);

    ms.destroy;
    ma.destroy;
    mi.destroy;
    mw.destroy;
    mf.destroy;
    msl.destroy;
    cc.destroy;
end.
