{  simg -- image processing library
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{*************************************************************************}
{******* CSImg_GetRaw ****************************************************}
{*************************************************************************}
{******* init ************************************************************}
constructor CSImg_GetRaw.init(cc:CSImg_Config; bn, nx,ny,nc:longint; fmt:TSImg_RawFormat);  { creates n(=:bn) buffers, ... }
var i : longint;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('CSImg_GetRaw.init: called with: bufnum='+int2str(bn)+' nx='+int2str(nx)+' ny='+int2str(ny)+' nc='+int2str(nc)+' fmt='+fmt2str(fmt));
    bufnum:=bn;
    if bufnum<1 then cc.error('CSImg_GetRaw.init: Invalid number of buffers');
    getmem2(bufs,sizeof(TSImg_RDImage)*bufnum);
    for i:=0 to bufnum-1 do
        SImg_New(cc, bufs[i], fmt, nx,ny,nc, true);
    SImg_DecDebugLevel;
end;

{******* close ***********************************************************}
procedure CSImg_GetRaw.close(cc:CSImg_Config);
var i : longint;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('CSImg_GetRaw.close');
    for i:=0 to bufnum-1 do
        SImg_Free(cc,bufs[i]);
    freemem2(bufs,sizeof(TSImg_RDImage)*bufnum);
    self.destroy;
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* CSImg_FileRead_Thread *******************************************}
{*************************************************************************}
type CSImg_FileRead_Thread = class(CSImg_Thread)
       constructor create(cc_:CSImg_Config; fr_:CSImg_FileRead);
       procedure execute; override;
      private
       fr : CSImg_FileRead;
       cc : CSImg_Config;
       end;

constructor CSImg_FileRead_Thread.create(cc_:CSImg_Config; fr_:CSImg_FileRead);
begin
    cc:=SImg_Config_Default.create(cc_);
    fr:=fr_;
    inherited create(false);
end;

procedure CSImg_FileRead_Thread.execute;
var i,j,idx : longint;
    d       : dword;
begin
    cc.debug('CSImg_FileRead_Thread.execute: Thread started');
    while (not terminated) do
    begin
        d:=$ffffffff;
        i:=-1;
        for j:=0 to fr.bufnum-1 do
            if fr.capt_id[j]<d then
            begin
                d:=fr.capt_id[j];
                i:=j;
            end;
        if i>=0 then
        begin
            idx:=0;
            while (idx<fr.bufs[i].dsize) and (not terminated) do
            begin
                j:=fpread(fr.fd, fr.bufs[i].ptb[idx], fr.bufs[i].dsize-idx);
                if j>0 then idx+=j;
                if j<fr.bufs[i].dsize-idx then delay(1);
            end;
            fr.capt_id[i]:=$ffffffff;
        end;
    end;
    cc.debug('CSImg_FileRead_Thread.execute: Thread stopped');
    cc.destroy;
end;


{*************************************************************************}
{******* CSImg_FileRead **************************************************}
{*************************************************************************}
{******* init ************************************************************}
constructor CSImg_FileRead.init(cc:CSImg_Config; const fn:shortstring; bn,nx,ny,nc:longint; fmt:TSImg_RawFormat; cct:boolean);  { fn: file name; ct: create capture thread }
var i : longint;
begin
    if cc=nil then cc:=SImg_Config_Default;
    inherited init(cc, bn, nx,ny,nc, fmt);
    getmem2(capt_id, sizeof(dword)*bufnum);
    for i:=0 to bufnum-1 do
    begin
        capt_id[i]:=$ffffffff;
    end;
    capt_cnt:=0;
    if cct then ct:=CSImg_FileRead_Thread.create(cc,self)
    else ct:=nil;
    if (fn='') or (fn='-')
    then fd:=fpopen('/dev/stdin',open_rdonly or open_nonblock)
    else fd:=fpopen(fn,open_rdonly or open_nonblock);
    if fd<0 then cc.error('CSImg_FileRead.init: '+fn+': Error opening file: '+StrError(fpGetErrno));
end;

{******* close ***********************************************************}
procedure CSImg_FileRead.close(cc:CSImg_Config);    { destroy ETimes entries, calls inherited }
begin
    if cc=nil then cc:=SImg_Config_Default;
    if (ct<>nil) and (ct is CSImg_FileRead_Thread) then
    begin
        ct.terminate;
        ct.destroy;
        ct:=nil;
    end;
    fpclose(fd);
    freemem2(capt_id, sizeof(dword)*bufnum);
    inherited close(cc);
end;

{******* ReadStart *******************************************************}
procedure CSImg_FileRead.ReadStart(cc:CSImg_Config; i:longint);
begin
    if cc=nil then cc:=SImg_Config_Default;
    if (i<0) or (i>=bufnum) then cc.error('CSImg_FileRead.ReadStart: Invalid buffer index');
    capt_id[i]:=capt_cnt;
    capt_cnt+=1;
end;

{******* ReadEnd *********************************************************}
procedure CSImg_FileRead.ReadEnd(cc:CSImg_Config; i:longint);
const timeout = 3;
var j,idx,cnt : longint;
    t         : double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if (i<0) or (i>=bufnum) then cc.error('CSImg_FileRead.ReadEnd: Invalid buffer index');
    if ct=nil then
    begin
        cnt:=0;
        idx:=0; j:=0;
        t:=stimer;
        while (idx<bufs[i].dsize) and (stimer-t<=timeout) and ((j>=0) or (fpGetErrno=0)) do
        begin
            if cnt>0 then delay(1);
            j:=fpread(fd,bufs[i].ptb[idx],bufs[i].dsize-idx);
            if j>0 then idx+=j;
            cnt+=1;
        end;
        if j<0 then cc.error('CSImg_FileRead.ReadEnd: Reading error'+': '+StrError(fpGetErrno))
        else if idx<bufs[i].dsize then cc.error('CSImg_FileRead.ReadEnd: Unable to read data (timeout)');
        capt_id[i]:=$ffffffff;
    end
    else begin
        t:=stimer;
        while (capt_id[i]<$ffffffff) and (stimer-t<timeout) do
            delay(1);
        if capt_id[i]<$ffffffff then cc.error('CSImg_FileRead.ReadEnd: Unable to read data (timeout)');
    end;
end;


{*************************************************************************}
{******* CSImg_RawSave ***************************************************}
{*************************************************************************}
procedure SImg_RawSave(cc:CSImg_Config; data:CSImg_GetRaw; const fo:shortstring; n:longint);
var i,j     : longint;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    SImg_DebugMessage('SImg_RawSave: data='+data.classname+' fo='+fo+' n='+int2str(n));
    i:=SImg_LastFileSetNum(fo)+1;
    if i<n then begin
        cc.warning('SImg_RawSave: '+format('Can only save %d images instead of %d',[i,n]));
        n:=i;
    end;

    for i:=0 to min(data.bufnum,n)-1 do
        data.ReadStart(cc,i);

    for i:=0 to n-1 do
    begin
        j:=i mod data.bufnum;
        data.ReadEnd(cc,j);
        SImg_WriteImageFile(cc, SImg_NumToSetFile(fo,i), data.bufs[j]);
        if i+data.bufnum<n then data.ReadStart(cc,j);
    end;

    data.close(cc);
    SImg_DecDebugLevel;
end;


