{  simg -- image processing library
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{*************************************************************************}
{******* SImg_BlendTile **************************************************}
{*************************************************************************}
{ copy tile into mosaic and blend borders
    ti          : tile info
    ft          : tile image
    fi          : destination (mosiac) image
    tx,tx       : tile position
}
procedure SImg_BlendTile(cc:CSImg_Config; ti:TSImg_TileInfo; ft,fi:TSImg_FPImage; tx,ty:longint);
var xa,ya,xz,yz,x,y,c,xb,yb     : longint;
    xa1,ya1,xz1,yz1             : longint;
    w,wy,rb                     : simgfp;
    d,e                         : ^simgfp;
begin
    if ft.pt=fi.pt then exit;   { only one tile }
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_BlendTile: ft='+SImg_InfoStr2(ft)+' fi='+SImg_InfoStr2(fi)+' tx='+int2str(tx)+' ty='+int2str(ty));

    { integrity checks }
    if not SImg_Recover(cc,fi) then cc.error('SImg_BlendTile: Invalid image: fi');
    if not SImg_Recover(cc,ft) then cc.error('SImg_BlendTile: Invalid image: ft');
    if (ti.nx<1) or (ti.ny<1) then cc.error('SImg_BlendTile: Invalid number of tiles');

    if (tx<0) or (tx>=ti.nx) then cc.error('SImg_BlendTile: Invalid tx');
    xa:=(tx*fi.nx) div ti.nx;
    xz:=((tx+1)*fi.nx) div ti.nx - 1;
    xb:=xz+1-xa;
    if tx>0 then xb+=ti.bb+ti.bu;
    if tx<ti.nx-1 then xb+=ti.bb+ti.bu;
    if (ty<0) or (ty>=ti.ny) then cc.error('SImg_BlendTile: Invalid ty');
    ya:=(ty*fi.ny) div ti.ny;
    yz:=((ty+1)*fi.ny) div ti.ny - 1;
    yb:=yz+1-ya;
    if ty>0 then yb+=ti.bb+ti.bu;
    if ty<ti.ny-1 then yb+=ti.bb+ti.bu;
    if (ft.nc<>fi.nc) or (ft.nx<>xb) or (ft.ny<>yb) then cc.error('SImg_BlendTile: Invalid dimensions: ti');

    { area of interest }
    if tx<=0 then
    begin
        xa1:=xa;                { first pixel at which wx=1 }
        xb:=0;                  { first pixel in tile }
    end
    else begin
        xa1:=xa+ti.bb;
        xa :=xa-ti.bb;
        xb:=ti.bu;
    end;
    if tx>=ti.nx-1 then xz1:=xz { last pixel at which wx=1 }
    else begin
        xz1:=xz-ti.bb;
        xz :=xz+ti.bb;
    end;

    if ty<=0 then
    begin
        ya1:=ya;                { first pixel at which wy=1 }
        yb:=0;                  { first pixel in tile }
    end
    else begin
        ya1:=ya+ti.bb;
        ya :=ya-ti.bb;
        yb:=ti.bu;
    end;
    if ty>=ti.ny-1 then yz1:=yz { last pixel at which wy=1 }
    else begin
        yz1:=yz-ti.bb;
        yz :=yz+ti.bb;
    end;

    if (xz1+1<=xa1-1) or (xz1+1<=xa1-1) then cc.error(cc.name+': Tile blending border (ti.bb) to large');

    { copying/blending }
    rb:=1/(1+2*ti.bb);
    for y:=ya to yz do
    begin
        if y<ya1 then wy:=1-(ya1-y)*rb
        else if y>yz1 then wy:=1-(y-yz1)*rb
        else wy:=1;
        d:=ft.pt + yb*ft.sl + xb*ft.nc; yb+=1;
        e:=fi.pt + y *fi.sl + xa*fi.nc;
        for x:=xa to xa1-1 do
        begin
            w:=wy*(1-(xa1-x)*rb);
            for c:=0 to ft.nc-1 do
                e[c]+=d[c]*w;
            d+=ft.nc;
            e+=fi.nc;
        end;
        for x:=xa1 to xz1 do
        begin
            for c:=0 to ft.nc-1 do
                e[c]+=d[c]*wy;
            d+=ft.nc;
            e+=fi.nc;
        end;
        for x:=xz1+1 to xz do
        begin
            w:=wy*(1-(x-xz1)*rb);
            for c:=0 to ft.nc-1 do
                e[c]+=d[c]*w;
            d+=ft.nc;
            e+=fi.nc;
        end;
    end;
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_CopyTile ***************************************************}
{*************************************************************************}
{ copy tile into mosaic and do not blend borders
    ti          : tile info
    ft          : file image
    fi          : destination (mosaic) image
    tx,tx       : tile position
}
procedure SImg_CopyTile(cc:CSImg_Config; ti:TSImg_TileInfo; rt,ri:TSImg_RDImage; tx,ty:longint);
var xa,ya,xz,yz,xb,yb,b : longint;
    d,e                 : ^byte;
begin
    if rt.pt=ri.pt then exit;   { only one tile }
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_CopyTile: rt='+SImg_InfoStr2(rt)+' ri='+SImg_InfoStr2(ri)+' tx='+int2str(tx)+' ty='+int2str(ty));

    { integrity checks }
    if not SImg_Recover(cc,ri) then cc.error('SImg_CopyTile: Invalid image: ri');
    if not SImg_Recover(cc,rt) or (rt.fmt<>ri.fmt) then cc.error('SImg_CopyTile: Invalid image: rt');
    if (tx<0) or (tx>=ti.nx) then cc.error('SImg_CopyTile: Invalid tx');
    xa:=(tx*ri.nx) div ti.nx;
    xz:=((tx+1)*ri.nx) div ti.nx - 1;
    xb:=xz+1-xa;
    if tx>0 then xb+=ti.bb+ti.bu;
    if tx<ti.nx-1 then xb+=ti.bb+ti.bu;
    if (ty<0) or (ty>=ti.ny) then cc.error('SImg_CopyTile: Invalid ty');
    ya:=(ty*ri.ny) div ti.ny;
    yz:=((ty+1)*ri.ny) div ti.ny - 1;
    yb:=yz+1-ya;
    if ty>0 then yb+=ti.bb+ti.bu;
    if ty<ti.ny-1 then yb+=ti.bb+ti.bu;
    if (rt.nc<>ri.nc) or (rt.nx<>xb) or (rt.ny<>yb) then cc.error('SImg_CopyTile: Invalid dimensions: rt');

    { without conversion ... }
    ri.bole:=rt.bole;
    ri.cf:=rt.cf;
    ri.co:=rt.co;

    { area of interest }
    if tx<=0 then xb:=0         { first pixel in tile }
    else xb:=ti.bu+ti.bb;
    if ty<=0 then yb:=0
    else yb:=ti.bu+ti.bb;

    { copy data }
    b:=rt.dsize div rt.size;    { bytes per pixel }
    xz:=b*rt.nc*(xz-xa+1);      { bytes to copy per row }
    d:=rt.ptb + (yb*rt.sl+xb*rt.nc)*b;
    e:=ri.ptb + (ya*ri.sl+xa*ri.nc)*b;
    for ya:=ya to yz do
    begin
        move(d^,e^,xz);
        d+=rt.sl*b;
        e+=ri.sl*b;
    end;
    SImg_DecDebugLevel;
end;

procedure SImg_CopyTile(cc:CSImg_Config; ti:TSImg_TileInfo; ft,fi:TSImg_FPImage; tx,ty:longint);
var rt,ri       : TSImg_RDImage;
begin
    if ft.pt=fi.pt then exit;   { only one tile }
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fi) then cc.error('SImg_CopyTile: Invalid image: fi');
    if not SImg_Recover(cc,ft) then cc.error('SImg_CopyTile: Invalid image: ft');
    ri.pt:=nil;  rt.pt:=nil;
    SImg_Convert(cc,fi,ri,true,true);
    SImg_Convert(cc,ft,rt,true,true);
    SImg_CopyTile(cc, ti, rt,ri, tx,ty);
end;


{*************************************************************************}
{******* SImg_GetTile ****************************************************}
{*************************************************************************}
{ Extract tile from mosaic
    ti          : tile info
    ft          : file image
    fi          : destination (mosaic) image
    tx,tx       : tile position
}
procedure SImg_GetTile(cc:CSImg_Config; ti:TSImg_TileInfo; ri:TSImg_RDImage; var rt:TSImg_RDImage; tx,ty:longint);
var ro                  : TSimg_RDImage;
    xa,ya,xz,yz,b       : longint;
    d,e                 : ^byte;
begin
    if ti.nx*ti.ny<=1 then
    begin
        SImg_Move(cc, ri,rt);
        exit;
    end;
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_GetTile: rt='+SImg_InfoStr2(rt)+' ri='+SImg_InfoStr2(ri)+' tx='+int2str(tx)+' ty='+int2str(ty));
    if not SImg_OverwriteAlert(ri,rt) then SImg_Free(cc,rt);
    if not SImg_Recover(cc,ri) then cc.error('SImg_GetTile: Invalid image: ri');

    { area of interest }
    if (tx<0) or (tx>=ti.nx) then cc.error('SImg_GetTile: Invalid tx');
    xa:=(tx*ri.nx) div ti.nx;
    xz:=((tx+1)*ri.nx) div ti.nx - 1;
    if tx>0 then xa-=ti.bb+ti.bu;
    if tx<ti.nx-1 then xz+=ti.bb+ti.bu;
    if (ty<0) or (ty>=ti.ny) then cc.error('SImg_GetTile: Invalid ty');
    ya:=(ty*ri.ny) div ti.ny;
    yz:=((ty+1)*ri.ny) div ti.ny - 1;
    if ty>0 then ya-=ti.bb+ti.bu;
    if ty<ti.ny-1 then yz+=ti.bb+ti.bu;
    if (xa<0) or (ya<0) or (xz>=ri.nx) or (yz>=ri.ny) then cc.error('SImg_GetTile: Invalid tile settings. Increase tile size or reduce border.');

    { copy data }
    ro.pt:=nil;
    SImg_New(cc, ro, ri.fmt, xz-xa+1,yz-ya+1,ri.nc, false);
    ro.bole:=ri.bole;
    ro.cf:=ri.cf;
    ro.co:=ri.co;
    b:=ri.dsize div ri.size;    { bytes per pixel }
    xz:=(xz-xa+1)*ri.nc*b;
    d:=ri.ptb + (ya*ri.sl+xa*ri.nc)*b;
    e:=ro.ptb;
    for ya:=ya to yz do
    begin
        move(d^,e^,xz);
        d+=ri.sl*b;
        e+=ro.sl*b;
    end;

    SImg_Move(cc, ro,rt);
    SImg_DecDebugLevel;
end;


procedure SImg_GetTile(cc:CSImg_Config; ti:TSImg_TileInfo; fi:TSImg_FPImage; var ft:TSImg_FPImage; tx,ty:longint);
var rt,ri       : TSImg_RDImage;
begin
    if ti.nx*ti.ny<=1 then
    begin
        SImg_Move(cc, fi,ft);
        exit;
    end;
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fi) then cc.error('SImg_GetTile: Invalid image: fi');
    if not SImg_OverwriteAlert(fi,ft) then SImg_Free(cc,ft);

    ri.pt:=nil;  rt.pt:=nil;
    SImg_Convert(cc,fi,ri,true,true);
    SImg_GetTile(cc, ti, ri,rt, tx,ty);
    SImg_Convert(cc, rt,ft,nil,true,false);
end;


